use std::{fs::create_dir_all, io::Cursor, path::PathBuf};

use directories::ProjectDirs;
use iced_video_player::Video;
use ::image::{DynamicImage, ImageReader};
use iced::{gradient, widget::container, Color};

use crate::{LauncherBackground, PossibleGames, VideoBackground, ORGANIZATION, PRODUCT};

use super::img_utils::round_image;

pub fn get_game_background(game: &PossibleGames, make_video: bool) -> Result<LauncherBackground, std::io::Error> {
    let proj_dirs = ProjectDirs::from("com", ORGANIZATION, PRODUCT).unwrap();
    let background_path = get_asset_file(&proj_dirs, game, "background")?;

    if background_path.extension().unwrap().to_str().unwrap() == "mp4" {
        let first_frame_path = get_asset_file(&proj_dirs, game, "first_frame_image")?;
        let first_frame = load_image_from_path(&first_frame_path)?;

        if !make_video {
            return Ok(LauncherBackground::Video(VideoBackground {
                first_frame,
                video: None
            }));
        }
        
        let video = Video::new(&url::Url::from_file_path(background_path).unwrap())
            .map_err(|err| {
                eprintln!("{err:#?}");
                std::io::Error::other("Video creation failed")
            })?;
            
        Ok(LauncherBackground::Video(VideoBackground {
            first_frame,
            video: Some(video)
        }))
    } else {
        let img = load_image_from_path(&background_path)?;
        Ok(LauncherBackground::Image(img))
    }
}

fn get_asset_file(proj_dirs: &ProjectDirs, game: &PossibleGames, identifier: &str) -> Result<PathBuf, std::io::Error> {
    let game_dir = match game {
        PossibleGames::WutheringWaves => proj_dirs.data_dir().join("kurogames\\wuwa"),
        PossibleGames::ZenlessZoneZero => proj_dirs.data_dir().join("hoyoverse\\zzz"),
        PossibleGames::HonkaiStarRail => proj_dirs.data_dir().join("hoyoverse\\hsr"),
        PossibleGames::GenshinImpact => proj_dirs.data_dir().join("hoyoverse\\gi"),
    };
    
    create_dir_all(&game_dir).map_err(|_| std::io::Error::new(
        std::io::ErrorKind::NotFound,
        format!("Failed to create game directory: {game_dir:?}")
    ))?;
    
    std::fs::read_dir(&game_dir)?
        .filter_map(std::result::Result::ok)
        .find(|entry| {
            entry.file_name()
                .to_string_lossy()
                .starts_with(&format!("{identifier}_"))
        })
        .map(|entry| entry.path())
        .ok_or_else(|| std::io::Error::new(
            std::io::ErrorKind::NotFound,
            format!("No {identifier} file found in {game_dir:?}")
        ))
}

fn load_image_from_path(path: &PathBuf) -> Result<DynamicImage, std::io::Error> {
    let data = std::fs::read(path)?;
    let cursor = Cursor::new(data);
    
    ImageReader::new(cursor)
        .with_guessed_format()
        .map_err(|_| std::io::Error::new(std::io::ErrorKind::InvalidData, "Invalid image format"))?
        .decode()
        .map_err(|_| std::io::Error::new(std::io::ErrorKind::InvalidData, "Failed to decode image"))
}

pub fn get_game_splash_dynamic_image(game: &PossibleGames) -> Option<DynamicImage> {
    let proj_dirs = ProjectDirs::from("com", ORGANIZATION, PRODUCT).unwrap();
    get_asset_file(&proj_dirs, game, "splash")
        .and_then(|path| load_image_from_path(&path))
        .ok()
}

pub fn get_game_icon_dynamic_image(game: &PossibleGames) -> DynamicImage {
    let file_data: &[u8] = if game == &PossibleGames::WutheringWaves { include_bytes!("../../resources/wutheringwaves-icon.png") } else {
        let proj_dirs = ProjectDirs::from("com", ORGANIZATION, PRODUCT).unwrap();
        &std::fs::read(get_asset_file(&proj_dirs, game, "icon").unwrap()).unwrap()
    };

    let cursor = Cursor::new(file_data);
    let img = ImageReader::new(cursor)
        .with_guessed_format()
        .unwrap()
        .decode()
        .unwrap()
        .resize(128, 128, ::image::imageops::FilterType::Lanczos3);

    round_image(img).unwrap()
}

fn rad(deg: f32) -> f32 {
    deg.to_radians()
}

pub fn style_container(direction: f32, use_gradient: bool) -> container::Style {
    let background = if use_gradient {
        Some(gradient::Linear::new(rad(direction))
            .add_stop(0.0, Color::from_rgba8(0, 0, 0, 0.0))
            .add_stop(0.75, Color::from_rgba8(0, 0, 0, 0.75))
            .into())
    } else {
        None
    };
    
    container::Style {
        text_color: Color::from_rgba8(255, 255, 255, 1.0).into(),
        background,
        ..container::Style::default()
    }
}